# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import warnings

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_powercycle_by_sleep(subsystem, nvme0):
    # sleep system for 10 seconds, to make DUT power off and on
    subsystem.power_cycle()


@pytest.mark.parametrize("nsid", [0, 1, 0xffffffff])
@pytest.mark.parametrize("stc", [1, 2])
def test_dst(nvme0, nsid, stc):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(stc, nsid).waitdone()

    # check dst log page till no dst in progress
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    while buf[0]:
        time.sleep(1)
        nvme0.getlogpage(0x6, buf, 32).waitdone()
        logging.info("current dst progress percentage: %d%%" % (buf[1] if(buf[1]) else 100))


@pytest.mark.parametrize("nsid", [2, 3, 8, 10, 0xff, 0xfffffffe])
@pytest.mark.parametrize("stc", [1, 2])
def test_dst_invalid_namespace(nvme0, nsid, stc):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.dst(stc, nsid).waitdone()


@pytest.mark.parametrize("nsid", [0, 1, 0xffffffff])
@pytest.mark.parametrize("stc", [1, 2])
def test_dst_in_progress(nvme0, nsid, stc):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(stc, nsid).waitdone()

    with pytest.warns(UserWarning, match="ERROR status: 01/1d"):
        nvme0.dst(stc, nsid).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 01/1d"):
        nvme0.dst(stc, nsid).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 01/1d"):
        nvme0.dst(stc, nsid).waitdone()

    # check dst log page till dst finishes
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    while buf[0]:
        time.sleep(1)
        nvme0.getlogpage(0x6, buf, 32).waitdone()
        logging.info("current dst progress percentage: %d%%" % (buf[1] if(buf[1]) else 100))


@pytest.mark.parametrize("nsid", [0, 1, 0xffffffff])
@pytest.mark.parametrize("stc", [1, 2])
def test_dst_abort(nvme0, nsid, stc):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(stc, nsid).waitdone()

    nvme0.dst(0xf, nsid).waitdone()

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


def test_dst_invalid_stc(nvme0, nsid=1):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.dst(0, nsid).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.dst(0xe, nsid).waitdone()
    nvme0.dst(0xf, nsid).waitdone()

    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


@pytest.mark.parametrize("stc", [1, 2])
def test_dst_abort_by_format(nvme0, nvme0n1, stc, nsid=1):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(stc, nsid).waitdone()

    nvme0.format(0, 0).waitdone()

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


def test_dst_short_abort_by_reset(nvme0):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(1, 0).waitdone()

    nvme0.reset()

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


def test_dst_extended_abort_by_reset(nvme0):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(2, 0).waitdone()

    nvme0.reset()

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert buf[0]

    nvme0.dst(0xf, 0).waitdone()
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


def test_pcie_reset_setup(pcie, nvme0):
    pcie.reset()
    nvme0.reset()


def test_dst_extended_abort_by_subsystem_reset(nvme0, subsystem):
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    nvme0.dst(2, 0).waitdone()

    time.sleep(2)
    subsystem.reset()
    nvme0.reset()

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert buf[0]

    nvme0.dst(0xf, 0).waitdone()
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]


@pytest.mark.parametrize("stc", [1, 2])
def test_dst_abort_by_sanitize(nvme0, nvme0n1, stc, nsid=1):
    if nvme0.id_data(331, 328) == 0:
        pytest.skip("sanitize operation is not supported")

    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    # aer callback function
    def cb(cdw0, status):
        warnings.warn("AER notification is triggered")
    nvme0.aer(cb)

    nvme0.dst(stc, nsid).waitdone()
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert buf[0]

    nvme0.sanitize().waitdone()  # sanitize command is completed

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()
        while buf.data(3, 2) & 0x7 != 1:  # sanitize operation is not completed
            nvme0.getlogpage(0x81, buf, 20).waitdone()
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)
            time.sleep(1)

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]
    vs = nvme0[8]
    logging.info("%d" %vs)
    if vs >= 0x010400:
        assert buf[4]&0xf == 0x09


@pytest.mark.parametrize("stc", [1, 2])
def test_dst_after_sanitize(nvme0, nvme0n1, stc, nsid=1):
    if nvme0.id_data(331, 328) == 0:
        pytest.skip("sanitize operation is not supported")

    # check dst
    buf = Buffer(4096)
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]

    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    nvme0.sanitize().waitdone()  # sanitize command is completed

    with pytest.warns(UserWarning, match="ERROR status: 00/1d"):
        # dst aborted due to in-progress sanitize
        nvme0.dst(stc, nsid).waitdone()

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()
        while buf.data(3, 2) & 0x7 != 1:  # sanitize operation is not completed
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)

    # check if dst aborted
    nvme0.getlogpage(0x6, buf, 32).waitdone()
    assert not buf[0]
    vs = nvme0[8]
    logging.info("%d" %vs)
    if vs >= 0x010400:
        assert buf[4]&0xf == 0x09



import time
import pytest
import logging

import nvme as d


def test_format(nvme0: d.Controller, nvme0n1):
    nvme0.format(0).waitdone()
    nvme0n1.format(512)


def test_download_firmware(nvme0):
    import PySimpleGUI as sg
    filename = sg.PopupGetFile('select the firmware binary file', 'pynvme')
    if filename:        
        logging.info("To download firmware binary file: " + filename)
        nvme0.downfw(filename)
    

def test_powercycle_by_sleep(subsystem, nvme0):
    # sleep system for 10 seconds, to make DUT power off and on
    subsystem.power_cycle()


def test_get_current_temperature(nvme0):
    from pytemperature import k2c
    smart_log = d.Buffer()
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.info("current temperature in SMART data: %0.2f degreeC" % k2c(ktemp))


def sg_show_hex_buffer(buf):
    import PySimpleGUI as sg
    layout = [ [sg.OK(), sg.Cancel()],
               [sg.Multiline(buf.dump(),
                             enter_submits=True,
                             disabled=True, 
                             size=(80, 25))]
             ]
    sg.Window(str(buf), layout, font=('monospace', 12)).Read()

    
def test_controller_identify_data(nvme0):
    b = d.Buffer(4096, "controller identify data")
    nvme0.identify(b).waitdone()
    sg_show_hex_buffer(b)


def test_namespace_identify_data(nvme0):
    b = d.Buffer(4096, "namespace identify data")
    nvme0.identify(b, 1, 0).waitdone()
    sg_show_hex_buffer(b)


def test_read_lba_data(nvme0, nvme0n1, qpair):
    import PySimpleGUI as sg
    
    lba = int(sg.PopupGetText("Which LBA to read?", "pynvme"))
    b = d.Buffer(512, "LBA 0x%08x" % lba)
    nvme0n1.read(qpair, b, lba).waitdone()
    sg_show_hex_buffer(b)

        
def test_get_dell_smart_attributes(nvme0):
    import PySimpleGUI as sg
    
    smart = d.Buffer()
    nvme0.getlogpage(0xCA, smart, 512).waitdone()

    l = []
    l.append('Byte |  Value  | Attribute')
    l.append('   0 |  %5d  | Re-Assigned Sector Count' % smart.data(0))
    l.append('   1 |  %5d  | Program Fail Count (Worst Case Component)' % smart.data(1))
    l.append('   2 |  %5d  | Program Fail Count (SSD Total)' % smart.data(2))
    l.append('   3 |  %5d  | Erase Fail Count (Worst Case Component)' % smart.data(3))
    l.append('   4 |  %5d  | Erase Fail Count (SSD Total)' % smart.data(4))
    l.append('   5 |  %5d  | Wear Leveling Count' % smart.data(5))
    l.append('   6 |  %5d  | Used Reserved Block Count (Worst Case Component)' % smart.data(6))
    l.append('   7 |  %5d  | Used Reserved Block Count (SSD Total)' % smart.data(7))
    l.append('11:8 |  %5d  | Reserved Block Count (SSD Total)' % smart.data(11, 8))

    layout = [[sg.Listbox(l, size=(70, 10))]]
    sg.Window("Dell SMART Attributes",
              layout+[[sg.OK()]],
              font=('monospace', 16)).Read()


def test_get_smart_health_information(nvme0):
    from pytemperature import k2c
    import PySimpleGUI as sg
    
    smart = d.Buffer()
    nvme0.getlogpage(0x02, smart, 512).waitdone()

    l = []
    l.append('  Byte |   Value   | Attribute')
    l.append('     0 |  %7d  | Critical Warning' % smart.data(0))
    l.append('  2: 1 |  %7d  | Composite Temperature (degree C)' % k2c(smart.data(2, 1)))
    l.append('     3 |  %7d  | Available Spare' % smart.data(3))
    l.append('     4 |  %7d  | Available Spare Threshold' % smart.data(4))
    l.append('     5 |  %7d  | Percentage Used' % smart.data(5))
    l.append('     6 |  %7d  | Endurance Group Critical Warning Summary' % smart.data(6))
    l.append(' 47:32 |%11d| Data Units Read (1000LBA)' % smart.data(47, 32))
    l.append(' 63:48 |%11d| Data Units Written (1000LBA)' % smart.data(63, 48))
    l.append(' 79:64 |%11d| Host Read Commands' % smart.data(79, 64))
    l.append(' 95:80 |%11d| Host Write Commands' % smart.data(95, 80))
    l.append('111:96 |  %7d  | Controller Busy Time (minutes)' % smart.data(111, 96))
    l.append('127:112|  %7d  | Power Cycles' % smart.data(127, 112))
    l.append('143:128|  %7d  | Power On Hours' % smart.data(143, 128))
    l.append('159:144|  %7d  | Unsafe Shutdowns' % smart.data(159, 144))
    l.append('175:160|  %7d  | Media and Data Integrity Errors' % smart.data(175, 160))
    l.append('191:176|  %7d  | Number of Error Information Log Entries' % smart.data(191, 176))
    l.append('195:192|  %7d  | Warning Composite Temperature Time (minutes)' % smart.data(195, 192))
    l.append('199:196|  %7d  | Critical Composite Temperature Time (minutes)' % smart.data(199, 196))
    l.append('201:200|  %7d  | Temperature Sensor 1 (degree C)' % k2c(smart.data(201, 200)))
    l.append('203:202|  %7d  | Temperature Sensor 2 (degree C)' % k2c(smart.data(203, 202)))
    l.append('205:204|  %7d  | Temperature Sensor 3 (degree C)' % k2c(smart.data(205, 204)))
    l.append('207:206|  %7d  | Temperature Sensor 4 (degree C)' % k2c(smart.data(207, 206)))
    l.append('209:208|  %7d  | Temperature Sensor 5 (degree C)' % k2c(smart.data(209, 208)))
    l.append('211:210|  %7d  | Temperature Sensor 6 (degree C)' % k2c(smart.data(211, 210)))
    l.append('213:212|  %7d  | Temperature Sensor 7 (degree C)' % k2c(smart.data(213, 212)))
    l.append('215:214|  %7d  | Temperature Sensor 8 (degree C)' % k2c(smart.data(215, 214)))
    l.append('219:216|  %7d  | Thermal Management Temperature 1 Transition Count' % smart.data(219, 216))
    l.append('223:220|  %7d  | Thermal Management Temperature 2 Transition Count' % smart.data(223, 220))
    l.append('227:224|  %7d  | Total Time For Thermal Management Temperature 1' % smart.data(227, 224))
    l.append('231:228|  %7d  | Total Time For Thermal Management Temperature 2' % smart.data(231, 228))

    layout = [[sg.Listbox(l, size=(70, 20))]]
    sg.Window("SMART/Health Information",
              layout+[[sg.OK()]],
              font=('monospace', 16)).Read()


def test_get_log_page(nvme0, lid=None):
    import PySimpleGUI as sg
    
    if lid == None:
        lid = int(sg.PopupGetText("Which Log ID to read?", "pynvme"))
    lbuf = d.Buffer(512, "%s, Log ID: %d" % (nvme0.id_data(63, 24, str), lid))
    nvme0.getlogpage(lid, lbuf).waitdone()
    sg_show_hex_buffer(lbuf)

    
def test_firmware_slot(nvme0, subsystem):
    import PySimpleGUI as sg
    
    filename = sg.PopupGetFile('select the firmware binary file', 'pynvme')
    if not filename:
        pytest.skip("no binary file found")

    def get_fw_slot(): #return: next reset, current
        buf = d.Buffer(512)
        nvme0.getlogpage(3, buf).waitdone()
        return buf[1], buf[0]
    
    # download slot 1
    nvme0.downfw(filename, 1)
    next_slot, this_slot = get_fw_slot()
    assert next_slot == 1

    # reset to activate
    subsystem.power_cycle()
    nvme0.reset()
    next_slot, this_slot = get_fw_slot()
    assert this_slot == 1
    assert next_slot == 0

    
def test_list_power_states(nvme0, buf):
    #according to NVMe Spec 1.4, Figure 247 and 248
    def print_ps(buf, ps):
        offset = 2048 + ps*32
        if buf.data(offset+1, offset+0):
            logging.info("PS %d" % ps)
            logging.info("===================================")
            
            ps_table = {0: 0, 0x40: 0.0001, 0x80: 0.01}
            mxps = 0.0001 if buf[offset+3]&1 else 0.01
            nops = 'Non-' if buf[offset+3]&2 else ''
            logging.info("Maximum Power: %.3f W" % (buf.data(offset+1, offset+0)*mxps))
            logging.info(nops+"Opertional State")

            logging.info("Entry Lantecy: %dus" %(buf.data(offset+7, offset+4)))
            logging.info("Exit Lantecy: %dus" %(buf.data(offset+11, offset+8)))

            logging.info("Relative Read Throughput: %d" %(buf.data(offset+12)))
            logging.info("Relative Read Latency: %d" %(buf.data(offset+13)))
            logging.info("Relative Write Throughput: %d" %(buf.data(offset+14)))
            logging.info("Relative Write Latency: %d" %(buf.data(offset+15)))

            ips = ps_table[buf[offset+18]]
            logging.info("Idle Power: %.3f W" % (buf.data(offset+17, offset+16)*ips))
            
            aps = ps_table[buf[offset+22]&0xc0]
            logging.info("Active Power: %.3f W with Workload %d" % (buf.data(offset+21, offset+20)*aps, buf[offset+22]&0x7))
            logging.info("")
        
    nvme0.identify(buf).waitdone()
    for ps in range(32):
        print_ps(buf, ps)



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_cq_p_phase_bit(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)

    # send commands
    sq[0] = SQE(4<<16+0, 1); sq.tail = 1
    sq[1] = SQE(3<<16+0, 1); sq.tail = 0
    sq[0] = SQE(2<<16+0, 1); sq.tail = 1
    sq[1] = SQE(1<<16+0, 1); sq.tail = 0

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10004
    assert cq[1][3] == 0x10003
    assert cq[2][3] == 0
    cq.head = 1
    time.sleep(0.1)
    assert cq[2][3] == 0x10002
    assert cq[0][3] == 0x10004
    cq.head = 2
    time.sleep(0.1)
    # p-bit changed to 0
    assert cq[0][3] == 0x00001


def test_cq_sqhd(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 2, PRP(), cqid=1)

    # send commands
    sq[0] = SQE(4<<16+0, 1); sq.tail = 1; time.sleep(0.1)
    sq[1] = SQE(3<<16+0, 1); sq.tail = 0; time.sleep(0.1)
    sq[0] = SQE(2<<16+0, 1); sq.tail = 1; time.sleep(0.1)
    sq[1] = SQE(1<<16+0, 1); sq.tail = 0; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    assert cq[0][2] == 0x10001
    assert cq[1][2] == 0x10000
    assert cq[2][2] == 0
    cq.head = 1
    time.sleep(0.1)
    assert cq[2][2] == 0x10001
    assert cq[0][2] == 0x10001
    cq.head = 2
    time.sleep(0.1)
    assert cq[0][2] == 0x10000
    assert cq[0][0] == 0
    assert cq[0][1] == 0


def test_p_invert_after_cq_2_pass(nvme0):
    """
    The value of the Phase Tag is inverted each pass filling the Complete Queue.
    """

    # cqid: 1, PC, depth: 2
    cq = IOCQ(nvme0, 1, 2, PRP())

    # create four SQ, both use the same CQ
    sq3 = IOSQ(nvme0, 3, 10, PRP(), cqid=1)

    # IO command templates: opcode and namespace
    write_cmd = SQE(1, 1)

    # write in sq3
    w1 = SQE(*write_cmd)
    write_buf = PRP(ptype=32, pvalue=0xaaaaaaaa)
    w1.prp1 = write_buf
    w1[10] = 1
    w1[12] = 1 # 0based
    w1.cid = 0x123
    sq3[0] = w1
    sq3.tail = 1

    # add some delay, so ssd should finish w1 before w2
    time.sleep(0.1)

    # write in sq5
    w1.cid = 0x567
    sq3[1] = w1
    sq3.tail = 2

    logging.info("aaa")
    # cqe for w1
    while CQE(cq[0]).p == 0: pass
    cqe = CQE(cq[0])
    assert cqe.cid == 0x123
    assert cqe.sqid == 3
    assert cqe.sqhd == 1
    cq.head = 1

    # cqe for w2
    logging.info("bbb")
    while CQE(cq[1]).p == 0: pass
    logging.info("ccc")
    cqe = CQE(cq[1])
    assert cqe.cid == 0x567
    assert cqe.sqid == 3
    assert cqe.sqhd == 2
    cq.head = 0

    assert CQE(cq[0]).p == 1
    assert CQE(cq[1]).p == 1

    # write in sq3
    w1.cid = 0x147
    sq3[2] = w1
    sq3.tail = 3

    # add some delay, so ssd should finish w1 before w2
    time.sleep(0.1)

    # write in sq5
    w1.cid = 0x167
    sq3[3] = w1
    sq3.tail = 4

    # cqe for w3
    while CQE(cq[0]).p == 1: pass
    cqe = CQE(cq[0])
    assert cqe.cid == 0x147
    assert cqe.sqid == 3
    assert cqe.sqhd == 3
    cq.head = 1

    # cqe for w4
    while CQE(cq[1]).p == 1: pass
    cqe = CQE(cq[1])
    assert cqe.cid == 0x167
    assert cqe.sqid == 3
    assert cqe.sqhd == 4
    cq.head = 0

    assert CQE(cq[0]).p == 0
    assert CQE(cq[1]).p == 0


def test_sq_cid1(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # send discontinuous cid commands
    sq[0] = SQE(4<<16+0, 1); sq.tail = 1; time.sleep(0.1)
    sq[1] = SQE(1<<16+0, 1); sq.tail = 2; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10004
    assert cq[1][3] == 0x10001
    sq.delete()
    cq.delete()

def test_sq_cid2(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # send max/min cid commands
    sq[0] = SQE(0xFFFF<<16+0, 1); 
    sq[1] = SQE(0<<16+0, 1); sq.tail = 2; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x1FFFF
    assert cq[1][3] == 0x10000
    sq.delete()
    cq.delete()

def test_cid_conflict(nvme0):
    mdts_lba = nvme0.mdts//512
    cq = IOCQ(nvme0, 1, 20, PRP())
    sq = IOSQ(nvme0, 1, 20, PRP(), cqid=1)

    # prp for the long buffer
    write_buf_1 = PRP(ptype=32, pvalue=0xaaaaaaaa)
    pages = mdts_lba//8
    pages -= 1

    prp_list = PRPList()
    prp_list_head = prp_list
    while pages:
        logging.info(pages)
        for i in range(63):
            if pages:
                prp_list[i] = PRP()
                pages -= 1
                logging.debug(pages)
        if pages>1:
            tmp = PRPList()
            prp_list[63] = tmp
            prp_list = tmp
            logging.debug("prp_list")
        elif pages==1:
            prp_list[63] = PRP()
            pages -= 1
            logging.debug(pages)

    #send first cmd   
    w1 = SQE((1<<16)+1, 1)
    w1.prp1 = write_buf_1
    w1.prp2 = prp_list_head
    w1[12] = mdts_lba-1 # 0based, nlba
    sq[0] = w1
    sq[1] = w1
    logging.info(sq[0])
    logging.info(sq[1])
    assert sq[0][0]>>16 == 1
    assert sq[1][0]>>16 == 1
    sq.tail = 2

    time.sleep(1)
    logging.info(cq[0])
    logging.info(cq[1])
    cqe = CQE(cq[0])
    assert cqe.p == 1
    assert (cq[0][3]>>17)&0x3ff == 0
    cqe = CQE(cq[1])
    assert cqe.p == 1
    assert (cq[0][3]>>17)&0x3ff == 0
    cq.head = 2

    sq.delete()
    cq.delete()   

def test_sq_rsv(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # Reserved field is non-zero.
    sq[0] = SQE((1<<16) + (7<<10) + 0, 1); sq.tail = 1; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10001
    sq.delete()
    cq.delete()

def test_sq_fuse_is_zero(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # FUSE field is zero.
    sq[0] = SQE((1<<16) + (0<<8), 1); sq.tail = 1; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    assert cq[0][3] == 0x10001
    sq.delete()
    cq.delete()

def test_sq_fuse_is_rsv(nvme0):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # FUSE field is 0x3(Reserved).
    sq[0] = SQE((1<<16) + (3<<8), 1); sq.tail = 1; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    #sct=0,sc=2(Invalid Field in Command)
    assert cq[0][3]>>17 == 0x0002
    sq.delete()
    cq.delete()
    
@pytest.mark.parametrize("opc_id", [0x3, 0x7, 0x0b, 0x0f, 0x12, 0x13, 0x16, 0x17, 0x1b])
def test_sq_opc_invalid_admin_cmd(nvme0,opc_id):
    #sct=0,sc=1(Invalid Command Opcode)
    with pytest.warns(UserWarning, match="ERROR status: 00/01"):
        nvme0.send_cmd(opc_id).waitdone()

@pytest.mark.parametrize("opc_id", [0x03, 0x07, 0x0a, 0x0b, 0x0f, 0x10, 0x12, 0x13, 0x14])
def test_sq_opc_invalid_nvm_cmd(nvme0,opc_id):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # OPC field is invalid.
    sq[0] = SQE((1<<16) + opc_id, 1); sq.tail = 1; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    #sct=0,sc=1(Invalid Command Opcode)
    assert (cq[0][3]>>17)&0x3ff == 0x0001

    sq.delete()
    cq.delete()

@pytest.mark.parametrize("ns_id", [0, 0x10, 0x100, 0x1000, 0xffffffff])
def test_sq_ns_invalid(nvme0,ns_id):
    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)

    # ns field is invalid.
    sq[0] = SQE(2, ns_id); sq.tail = 1; time.sleep(0.1)

    # check cq
    time.sleep(0.1)
    #sct=0,sc=0x0b(Invalid Namespace or Format)
    assert (cq[0][3]>>17)&0x3ff == 0x000b

    sq.delete()
    cq.delete()    



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


def test_controller_cap(nvme0):
    logging.info("cap: 0x%lx" % nvme0.cap)

    #verify mpsmax and mpsmin
    mps_min = (nvme0.cap>>48) & 0xf
    mps_max = (nvme0.cap>>52) & 0xf
    mps_max_size=2**(12+mps_max)
    mps_min_size=2**(12+mps_min)
    logging.info("mps_min_size :{}".format(mps_min_size))
    logging.info("mps_max_size :{}".format(mps_max_size))
    assert mps_max >= mps_min

    #verify cc.mps is smaller than mpsmin and laeger than mpsmax
    cc=nvme0[0x14]
    cc_mps= nvme0[0x14]>>6&0xf
    cc_mps_size=2**(12+cc_mps)
    logging.info("cc_mps_size:{}".format(cc_mps_size))
    assert cc_mps_size <=mps_max_size
    assert cc_mps_size >=mps_min_size

    css = (nvme0.cap>>37) & 0xff
    assert css == 1

    nssrs=nvme0.cap>>36&0x1
    assert nssrs==1


def test_controller_version(nvme0):
    logging.info("ver: 0x%x" % nvme0[8])
    assert (nvme0[8]>>16) == 1


def test_controller_cc(nvme0):
    logging.info("cc: 0x%x" % nvme0[0x14])
    assert (nvme0[0x14]>>16) == 0x46


def test_controller_reserved(nvme0):
    assert nvme0[0x18] == 0
    nvme0[0x18]=0x1111
    assert nvme0[0x18]==0


def test_controller_csts(nvme0):
    logging.info("csts: 0x%x" % nvme0[0x1c])
    assert nvme0[0x1c]&1 == 1


def test_controller_cap_to(nvme0):
    logging.info("cc :{}".format(nvme0[0x14]))
    logging.info("cap timeout :{}".format(nvme0.cap>>24&0xff))
    timeout=nvme0.cap>>24&0xff

    #change cc.en from '1' to '0'
    nvme0[0x14] = 0

    time_start=int(time.time())
    max_time=timeout *500 //1000

    #wait csts.rdy change from '1' to '0'
    while not (nvme0[0x1c]&0x1)==0:
        assert int(time.time())-time_start <= max_time


    if 0 != nvme0.init_adminq():
        raise NvmeEnumerateError("fail to init admin queue")

    #change cc.en from '0' to '1'
    nvme0[0x14] = 0x00460001

    #wait csts.rdy change from '0' to '1'
    time_start=int(time.time())
    while not (nvme0[0x1c]&0x1)==1:
        assert int(time.time())-time_start <= max_time



def test_controller_cap_mqes(nvme0):
    mqes=nvme0.cap&0xffff
    logging.info("mqes:{}".format(mqes))
    assert mqes >=1

    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        cq = IOCQ(nvme0, 1, mqes+5, PRP())

    cq = IOCQ(nvme0, 1, mqes, PRP())
    with pytest.warns(UserWarning, match="ERROR status: 01/02"):
        sq = IOSQ(nvme0, 1, mqes+5, PRP(), cqid=1)


def test_controller_ams(nvme0):
    ams=nvme0.cap>>17&0x3
    logging.info("AMS:{}".format(ams))


def test_controller_intms_and_intmc(nvme0):
    intms=nvme0[0xc]
    intmc=nvme0[0x10]
    logging.info("intms:{},intmc:{}".format(intms,intmc))
    nvme0[0xc]=0x0
    nvme0[0x10]=0x0
    #verify intms and intmc do not change value
    assert nvme0[0xc]==intms
    assert nvme0[0x10]==intmc


def test_controller_cc_iocqes(nvme0):
    iocqes=nvme0[0x14]>>20&0xf
    logging.info("iocqes:{}".format(iocqes))

    cqes=nvme0.id_data(513,513)
    logging.info("CQES:{}".format(cqes))

    cqes_max=cqes>>4&0xf
    cqes_min=cqes &0xf
    logging.info("cqes_max:{},cqes_min:{}".format(cqes_max,cqes_min))
    assert cqes_max >= cqes_min
    assert cqes_min == 4

    assert iocqes >= cqes_min
    assert iocqes <= cqes_max

def test_controller_cc_iosqes(nvme0):
    iosqes=nvme0[0x14]>>16&0xf
    logging.info("iosqes:{}".format(iosqes))

    sqes=nvme0.id_data(512,512)
    logging.info("SQES:{}".format(sqes))

    sqes_max=sqes>>4&0xf
    sqes_min=sqes &0xf
    logging.info("sqes_max:{},sqes_min:{}".format(sqes_max,sqes_min))
    assert sqes_max >= sqes_min
    assert sqes_min == 6

    assert iosqes >= sqes_min
    assert iosqes <= sqes_max


def test_controller_cc_shn(nvme0,subsystem):
    cc=nvme0[0x14]
    logging.info("cc:{}".format(cc))
    shn=nvme0[0x14]>>14&0x3
    logging.info("shn:{}".format(shn))

    begin_time=time.time()
    subsystem.shutdown_notify()
    end_time=time.time()
    time_spend_normal=end_time-begin_time
    logging.info("normal shn ,time spend is {}".format(time_spend_normal))

    nvme0.reset()


def test_controller_cc_en(nvme0):
    en=nvme0[0x14]&0x1
    logging.info("en:{}".format(en))
    logging.info("cc:{}".format(nvme0[0x14]))
    assert nvme0[0x14] == 0x00460001

    cq = IOCQ(nvme0, 1, 3, PRP())
    sq = IOSQ(nvme0, 1, 3, PRP(), cqid=1)
    sq[0] = SQE(2, 1); sq.tail = 1; time.sleep(0.1)
    # check cq
    time.sleep(0.1)
    assert (cq[0][3]>>17)&0x3ff == 0x0000

    sq.delete()
    cq.delete()

    #change cc.en from '1' to '0'
    nvme0[0x14] = 0

    #wait csts.rdy change from '1' to '0'
    while not (nvme0[0x1c]&0x1)==0:
        pass

    nvme0.aer()
    try:
        nvme0.waitdone()
    except Exception as e:
        logging.warning(e)


    if 0 != nvme0.init_adminq():
        raise NvmeEnumerateError("fail to init admin queue")

    #change cc.en from '0' to '1'
    nvme0[0x14] = 0x00460001

    #wait csts.rdy change from '0' to '1'
    while not (nvme0[0x1c]&0x1)==1:
        pass

    assert nvme0[0x14] ==0x00460001


def test_controller_cc_css(nvme0):
    css=nvme0[0x14]>>4&0x7
    logging.info("css:{}".format(css))


def test_controller_mdts(nvme0,nvme0n1):
    mps_min = (nvme0.cap>>48) & 0xf
    logging.info("mps_min:{}".format(mps_min))

    mdts=nvme0.id_data(77,77)
    logging.info("mdts:{}".format(mdts))
    max_data_size=(2**mdts)*(2**(12+mps_min))
    logging.info(max_data_size)
    max_pages=max_data_size//4//1024
    pages=max_pages-1
    max_lba=max_data_size//512

    cq = IOCQ(nvme0, 1, 5, PRP())
    sq = IOSQ(nvme0, 1, 5, PRP(), cqid=1)

    # prp for the long buffer
    write_buf_1 = PRP(ptype=32, pvalue=0xaaaaaaaa)
    prp_list = PRPList()
    prp_list_head = prp_list
    while pages:
        for i in range(511):
            if pages:
                prp_list[i] = PRP()
                pages -= 1
                logging.debug(pages)

        if pages>1:
            tmp = PRPList()
            prp_list[511] = tmp
            prp_list = tmp
            logging.debug("prp_list")

        elif pages==1:
            prp_list[511] = PRP()
            pages -= 1
            logging.debug(pages)

    w1 = SQE(1, 1)
    w1.prp1 = write_buf_1
    w1.prp2 = prp_list_head
    w1[12] = max_lba-1 # 0based, nlba
    w1.cid = 0x123
    sq[0] = w1
    sq.tail = 1

    time.sleep(1)
    cqe = CQE(cq[0])
    logging.info(cqe)
    assert cqe.p == 1
    assert cqe.cid == 0x123
    assert cqe.sqhd == 1
    assert cqe.status == 0
    cq.head = 1

    w2 = SQE(1, 1)
    w2.prp1 = write_buf_1
    w2.prp2 = prp_list_head
    w2[12] = max_lba # 0based, nlba
    w2.cid = 0x234
    sq[1] = w2
    sq.tail = 2

    time.sleep(1)
    cqe = CQE(cq[1])
    logging.info(cqe)
    assert cqe.p == 1
    assert cqe.cid == 0x234
    assert cqe.sqhd == 2
    assert cqe.status == 2
    cq.head = 2

    r1 = SQE(2, 1)
    r1.prp1 = write_buf_1
    r1.prp2 = prp_list_head
    r1[12] = max_lba # 0based, nlba
    r1.cid = 0x345
    sq[2] = r1
    sq.tail = 3

    time.sleep(1)
    cqe = CQE(cq[2])
    logging.info(cqe)
    assert cqe.p == 1
    assert cqe.cid == 0x345
    assert cqe.sqhd == 3
    assert cqe.status == 2
    cq.head = 3

    r2 = SQE(2, 1)
    r2.prp1 = write_buf_1
    r2.prp2 = prp_list_head
    r2[12] = max_lba-1 # 0based, nlba
    r2.cid = 0x456
    sq[3] = r2
    sq.tail = 4

    time.sleep(1)
    cqe = CQE(cq[3])
    logging.info(cqe)
    assert cqe.p == 1
    assert cqe.cid == 0x456
    assert cqe.sqhd == 4
    assert cqe.status == 0
    cq.head = 4

    sq.delete()
    cq.delete()



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_getlogpage_page_id(nvme0, buf):
    for lid in (1, 2, 3):
        nvme0.getlogpage(lid, buf).waitdone()

    for lid in (0, 0x6f, 0x7f, 0xff):
        with pytest.warns(UserWarning, match="ERROR status: 01/09"):
            nvme0.getlogpage(lid, buf).waitdone()
        

@pytest.mark.parametrize("repeat", range(32))
def test_getlogpage_invalid_numd(nvme0, repeat):
    dts = nvme0.mdts//4 + 1 + repeat
    buf = Buffer(dts*4)

    for lid in (1, 2, 3):
        with pytest.warns(UserWarning, match="ERROR status: 00/02"):
            nvme0.getlogpage(lid, buf).waitdone()


def test_getlogpage_after_error(nvme0, nvme0n1, buf, qpair):
    nvme0.getlogpage(1, buf).waitdone()
    assert buf.data(7, 0) == 0
    nvme0n1.write_uncorrectable(qpair, 0, 8).waitdone()

    # generate 2 errors
    with pytest.warns(UserWarning, match="ERROR status: 02/81"):
        nvme0n1.read(qpair, buf, 0, 8).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/81"):
        nvme0n1.read(qpair, buf, 0, 8).waitdone()

    time.sleep(0.1) # wait error information ready
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(1, buf).waitdone()
    nerror1 = buf.data(7, 0)
    nerror2 = buf.data(64+7, 64)
    assert nerror1 == nerror2+1

    nvme0n1.write(qpair, buf, 0, 8).waitdone()
    

@pytest.mark.parametrize("len", (1, 2, 4))
def test_getlogpage_data_unit_read(nvme0, nvme0n1, buf, qpair, len):
    if not nvme0n1.supports(5):
        pytest.skip("compare is not support")

    nvme0n1.write(qpair, buf, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nread1 = buf.data(47, 32)

    # read
    for i in range(1000):
        nvme0n1.read(qpair, buf, 0, len).waitdone()
    nvme0.getlogpage(2, buf).waitdone()
    nread2 = buf.data(47, 32)
    assert nread2 == nread1+len

    # compare
    nvme0n1.read(qpair, buf, 0, len).waitdone()  # get correct data
    for i in range(1000):
        nvme0n1.compare(qpair, buf, 0, len).waitdone()
    nvme0.getlogpage(2, buf).waitdone()
    nread3 = buf.data(47, 32)
    assert nread3 == nread2+len

    # verify
    if nvme0n1.supports(0xc):
        for i in range(1000):
            nvme0n1.verify(qpair, 0, len).waitdone()
        nvme0.getlogpage(2, buf).waitdone()
        nread4 = buf.data(47, 32)
        assert nread4 == nread3+len    


@pytest.mark.parametrize("len", (1, 2, 4))
def test_getlogpage_data_unit_write(nvme0, nvme0n1, len, buf, qpair):
    if not nvme0n1.supports(5):
        pytest.skip("compare is not support")

    nvme0n1.write(qpair, buf, 0).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite1 = buf.data(63, 48)

    for i in range(1000):
        nvme0n1.write(qpair, buf, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite2 = buf.data(63, 48)
    assert nwrite2 == nwrite1+len
    
    for i in range(1000):
        nvme0n1.write_uncorrectable(qpair, 0, len).waitdone()

    nvme0.getlogpage(2, buf).waitdone()
    nwrite3 = buf.data(63, 48)
    assert nwrite3 == nwrite2

    nvme0n1.write(qpair, Buffer(4096), 0, 8).waitdone()
    

def test_getlogpage_power_cycle_count(nvme0, subsystem, buf):
    def get_power_cycles(nvme0):
        nvme0.getlogpage(2, buf, 512).waitdone()
        return buf.data(115, 112)

    import time
    start_time = time.time()
    powercycle = get_power_cycles(nvme0)

    subsystem.power_cycle(10)
    nvme0.reset()
    assert get_power_cycles(nvme0) == powercycle+1


def test_getlogpage_namespace(nvme0, buf):
    nvme0.getlogpage(2, buf, nsid=1).waitdone()
    nvme0.getlogpage(2, buf, nsid=0xffffffff).waitdone()

    # getlogpage for id=2 nsid can be 0 or 0xffffffff
    nvme0.getlogpage(2, buf, nsid=0).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.getlogpage(2, buf, nsid=2).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/0b"):
        nvme0.getlogpage(2, buf, nsid=0xfffffffe).waitdone()

        
def test_getlogpage_smart_composite_temperature(nvme0):
    smart_log = Buffer()
    
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.debug("temperature: %d degreeF" % ktemp)

    # warning with AER
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        # over composite temperature threshold
        nvme0.setfeatures(4, cdw11=ktemp-10).waitdone()
        
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()
        logging.debug("0x%x" % smart_log.data(0))
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()    
        assert smart_log.data(0) & 0x2

        # higher threshold
        nvme0.setfeatures(4, cdw11=ktemp+10).waitdone()

    # aer is not expected
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.debug("temperature: %d degreeF" % ktemp)

    # revert to default
    orig_config = 0
    def getfeatures_cb_4(cdw0, status):
        nonlocal orig_config; orig_config = cdw0
    nvme0.getfeatures(4, sel=1, cb=getfeatures_cb_4).waitdone()
    nvme0.setfeatures(4, cdw11=orig_config).waitdone()    

    
def test_getlogpage_persistent_event_log(nvme0):
    if not (nvme0.id_data(261)&0x10):
        pytest.skip("feature sv is not supported")


def test_getlogpage_firmware_slot_info_nsid_1(nvme0, buf):
    nvme0.getlogpage(3, buf, 512, nsid=0).waitdone()
    nvme0.getlogpage(3, buf, 512, nsid=0xffffffff).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0.getlogpage(3, buf, 512, nsid=1).waitdone()



#
#  BSD LICENSE
#
#  Copyright (c) Crane Chu <cranechu@gmail.com>
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#    * Neither the name of Intel Corporation nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# -*- coding: utf-8 -*-


import time
import pytest
import random
import logging
import inspect

import nvme as d


def pytest_addoption(parser):
    parser.addoption(
        "--pciaddr", action="store", default="", help="pci (BDF) address of the device under test, e.g.: 02:00.0"
    )


@pytest.fixture(scope="function", autouse=True)
def script(request):
    # skip empty tests
    sourcecode = inspect.getsourcelines(request.function)[0]
    if 'pass' in sourcecode[-1] and len(sourcecode) < 5:
        pytest.skip("empty test function")

    # measure test time, and set random seed by time
    start_time = time.time()
    d.srand(int(start_time*1000000)&0xffffffff)
    yield
    logging.info("test duration: %.3f sec" % (time.time()-start_time))


@pytest.fixture(scope="session")
def pciaddr(request):
    return request.config.getoption("--pciaddr")


@pytest.fixture(scope="function")
def pcie(pciaddr):
    ret = d.Pcie(pciaddr)
    yield ret
    ret.close()

    
@pytest.fixture(scope="function")
def nvme0(pcie):
    ret = d.Controller(pcie)
    yield ret


@pytest.fixture(scope="function")
def subsystem(nvme0):
    ret = d.Subsystem(nvme0)
    yield ret


@pytest.fixture(scope="function")
def nvme0n1(nvme0):
    ret = d.Namespace(nvme0)
    yield ret
    ret.close()


@pytest.fixture(scope="function")
def qpair(nvme0):
    num_of_entry = (nvme0.cap & 0xffff) + 1
    num_of_entry = min(1024, num_of_entry)
    ret = d.Qpair(nvme0, num_of_entry)
    yield ret
    ret.delete()

    
@pytest.fixture(scope="function")
def tcg(nvme0):
    ret = d.Tcg(nvme0)
    yield ret
    ret.close()


@pytest.fixture(scope="session")
def buf():
    ret = d.Buffer(4096, "pynvme buffer")
    yield ret
    del ret


@pytest.fixture(scope="function")
def verify(nvme0n1):
    ret = nvme0n1.verify_enable(True)
    yield ret


@pytest.fixture(scope="function")
def aer():
    assert False, "aer fixture is replaced by admin command nvme0.aer()"


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()
    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "rep_" + rep.when, rep)



import time
import pytest
import logging

import nvme as d


def test_precondition_format(pcie, nvme0, nvme0n1, subsystem):
    pcie.reset()
    nvme0.reset()
    subsystem.power_cycle()
    nvme0.reset()
    nvme0n1.format(512)
    

def test_quarch_dirty_power_cycle_single(nvme0, poweron=None, poweroff=None):
    region_end = 256*1000*1000  # 1GB
    qdepth = min(1024, 1+(nvme0.cap&0xffff))
    
    # get the unsafe shutdown count
    def power_cycle_count():
        buf = d.Buffer(4096)
        nvme0.getlogpage(2, buf, 512).waitdone()
        return buf.data(115, 112)
    
    # run the test one by one
    subsystem = d.Subsystem(nvme0, poweron, poweroff)
    nvme0n1 = d.Namespace(nvme0, 1, region_end)
    assert True == nvme0n1.verify_enable(True)
    orig_unsafe_count = power_cycle_count()
    logging.info("power cycle count: %d" % orig_unsafe_count)

    # 128K random write
    cmdlog_list = [None]*1000
    with nvme0n1.ioworker(io_size=256,
                          lba_random=True,
                          read_percentage=30,
                          region_end=region_end,
                          time=30,
                          qdepth=qdepth, 
                          output_cmdlog_list=cmdlog_list):
        # sudden power loss before the ioworker end
        time.sleep(10)
        subsystem.poweroff()

    # power on and reset controller
    time.sleep(5)
    subsystem.poweron()
    nvme0.reset()

    # verify data in cmdlog_list
    logging.info(cmdlog_list[-10:])
    read_buf = d.Buffer(256*512)
    qpair = d.Qpair(nvme0, 10)
    for cmd in cmdlog_list:
        slba = cmd[0]
        nlba = cmd[1]
        op = cmd[2]
        if nlba and op==1:
            def read_cb(cdw0, status1):
                nonlocal slba
                if status1>>1:
                    logging.info("slba 0x%x, status 0x%x" % (slba, status1>>1))
            #logging.info("verify slba 0x%x, nlba %d" % (slba, nlba))
            nvme0n1.read(qpair, read_buf, slba, nlba, cb=read_cb).waitdone()
            # re-write to clear CRC mismatch
            nvme0n1.write(qpair, read_buf, slba, nlba, cb=read_cb).waitdone()
    qpair.delete()
    nvme0n1.close()

    # verify unsafe shutdown count
    unsafe_count = power_cycle_count()
    logging.info("power cycle count: %d" % unsafe_count)
    assert unsafe_count == orig_unsafe_count+1

    
# define the power on/off funciton
class quarch_power:
    def __init__(self, url: str, event: str, port: int):
        self.url = url
        self.event = event
        self.port = port
        
    def __call__(self):
        import quarchpy
        logging.debug("power %s by quarch device %s on port %d" %
                      (self.event, self.url, self.port))
        pwr = quarchpy.quarchDevice(self.url)
        
        if self.port == None:
            # serial port
            if self.event == "down":
                # cut down power with data link at the same time
                pwr.sendCommand("signal:all:source 7")
            pwr.sendCommand("run:power %s" % self.event)
        else:
            # network 4-port
            if self.event == "down":
                # cut down power with data link at the same time
                pwr.sendCommand("signal:all:source 7 <%d>" % self.port)
            pwr.sendCommand("run:power %s <%d>" % (self.event, self.port))
            
        pwr.closeConnection()
        

@pytest.mark.parametrize("repeat", range(10))
def test_quarch_dirty_power_cycle_multiple(pciaddr, nvme0, repeat):
    device_list = {
        "0000:06:00.0": # pynvme's software-defined power cycle
        (None,
         None),  
        "0000:3d:00.0": # pynvme's software-defined power cycle
        (None,
         None),  
        "0000:08:00.0":
        (quarch_power("SERIAL:/dev/ttyUSB0", "up", None),
         quarch_power("SERIAL:/dev/ttyUSB0", "down", None)),
        "0000:01:00.0":
        (quarch_power("REST:192.168.1.11", "up", 4),
         quarch_power("REST:192.168.1.11", "down", 4)),
        "0000:55:00.0":
        (quarch_power("REST:192.168.1.11", "up", 1),
         quarch_power("REST:192.168.1.11", "down", 1)),
        "0000:51:00.0":
        (quarch_power("REST:192.168.1.11", "up", 3),
         quarch_power("REST:192.168.1.11", "down", 3)),
    }

    poweron, poweroff = device_list[pciaddr]
    test_quarch_dirty_power_cycle_single(nvme0, poweron, poweroff)




# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import pytest
import logging

import nvme as d


def mi_vpd_write(nvme0, data, offset=0, length=256):
    nvme_status = 0
    mi_status = 0
    response = 0
    def mi_send_cb(dword0, status1):
        nonlocal nvme_status
        nonlocal mi_status
        nonlocal response
        nvme_status = (status1 >> 1)
        mi_status = (dword0 & 0xff)
        response = (dword0 >> 8)
    nvme0.mi_send(6, offset, length, data, cb=mi_send_cb).waitdone()
    assert nvme_status == 0
    return mi_status, response

def mi_vpd_read(nvme0, data, offset=0, length=256):
    nvme_status = 0
    mi_status = 0
    response = 0
    def mi_receive_cb(dword0, status1):
        nonlocal nvme_status
        nonlocal mi_status
        nonlocal response
        nvme_status = (status1 >> 1)
        mi_status = (dword0 & 0xff)
        response = (dword0 >> 8)
    nvme0.mi_receive(5, offset, length, data, cb=mi_receive_cb).waitdone()
    assert nvme_status == 0
    return mi_status, response

    
def test_vpd_write_and_read(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    write_buf = d.Buffer(256, pvalue=100, ptype=0xbeef)
    read_buf = d.Buffer(256)
    
    status, response = mi_vpd_write(nvme0, write_buf)
    assert status == 0
    assert response == 0
    
    status, response = mi_vpd_read(nvme0, read_buf)
    assert status == 0
    assert response == 0

    assert write_buf != read_buf
    assert write_buf[:] == read_buf[:]


def test_reset(nvme0, subsystem):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    nvme0.mi_send(7, 0).waitdone()
    nvme0.reset()


def test_invalid_operation(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    dword0 = nvme0.mi_send(0xbf).waitdone()
    assert dword0&0xff == 0x3


def test_configuration_get_invalid(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    dword0 = nvme0.mi_send(4, 1).waitdone()
    logging.info(hex(dword0))

    
def test_configuration_get_health_status_change(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    dword0 = nvme0.mi_send(4, 2).waitdone()
    logging.info(hex(dword0))


def test_configuration_set_health_status_change(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    dword0 = nvme0.mi_send(3, 2, 0).waitdone()
    logging.info(hex(dword0))


def test_read_nvme_mi_data_structure_nvm_subsystem_information(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    buf = d.Buffer(0x2000)
    dword0 = nvme0.mi_receive(0, 0, 0, buf).waitdone()
    logging.info(hex(dword0))
    logging.info(buf.dump(3))

# use mi_send to read nvme mi data structure: illegal
def test_read_nvme_mi_data_structure_nvm_subsystem_information_wrong_command(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    buf = d.Buffer(0x2000)
    dword0 = nvme0.mi_send(0, 0, 0, buf).waitdone()
    logging.info(hex(dword0))
    logging.info(buf.dump(3))
    

def test_read_nvme_mi_data_structure_port_information(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    buf = d.Buffer(0x2000)
    dword0 = nvme0.mi_receive(0, 1<<24, 0, buf).waitdone()
    logging.info(hex(dword0))
    logging.info(buf.dump(32))
    assert buf[0] == 1

    
def test_read_nvme_mi_data_structure_port_information_wrong_port(nvme0):
    if not nvme0.supports(0x1d) or not nvme0.supports(0x1e):
        pytest.skip("mi commands are not supported")
        
    buf = d.Buffer(0x2000)
    dword0 = nvme0.mi_receive(0, (1<<24)+(1<<16), 0, buf).waitdone()
    logging.info(hex(dword0))
    logging.info(buf.dump(32))
    



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture()
def cq(nvme0):
    ret = IOCQ(nvme0, 1, 10, PRP())
    yield ret
    ret.delete()

@pytest.fixture()
def sq(nvme0, cq):
    ret = IOSQ(nvme0, 1, 10, PRP(), cq.id)
    yield ret
    ret.delete()

    
def test_write_large_lba(nvme0, nvme0n1, buf, qpair):
    ncap = nvme0n1.id_data(15, 8)
    
    nvme0n1.write(qpair, buf, ncap-1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, ncap).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, ncap+1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, ncap-1, 2).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, 0xffffffff00000000).waitdone()


@pytest.mark.parametrize("repeat", range(32))
def test_write_max_namespace_size(nvme0, nvme0n1, buf, repeat, qpair):
    nsze = nvme0n1.id_data(7, 0)
    ncap = nvme0n1.id_data(15, 8)
    assert nsze == ncap
    
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, nsze+repeat).waitdone()

    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, 0xffffffff+repeat).waitdone()
    
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write(qpair, buf, 0xffffffffffffffff-repeat).waitdone()
    

def test_write_fua(nvme0, nvme0n1, buf, qpair):
    for i in range(100):
        # write with FUA enabled
        nvme0n1.write(qpair, buf, 0, 8, 1<<30).waitdone()


def test_write_bad_number_blocks(nvme0, nvme0n1, qpair):
    mdts = nvme0.mdts//512
    buf = Buffer(mdts*512+4096)

    nvme0n1.write(qpair, buf, 0, mdts-1).waitdone()
    nvme0n1.write(qpair, buf, 0, mdts).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0n1.write(qpair, buf, 0, mdts+1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/02"):
        nvme0n1.write(qpair, buf, 0, mdts+2).waitdone()

    nlb = 1
    while nlb < mdts:
        nvme0n1.write(qpair, buf, 0, nlb).waitdone()
        nlb = nlb<<1

        
@pytest.mark.parametrize("ioflag", [0, 0x4000, 0x8000, 0xc000])
def test_write_valid(nvme0, nvme0n1, ioflag, qpair):
    # prepare data buffer and IO queue
    read_buf = Buffer(512)
    write_buf = Buffer(512)
    write_buf[10:21] = b'hello world'

    # send write and read command
    def write_cb(cdw0, status1):  # command callback function
        nvme0n1.read(qpair, read_buf, 0, 1)
    nvme0n1.write(qpair, write_buf, 0, 1, io_flags=ioflag, cb=write_cb)

    # wait commands complete and verify data
    assert read_buf[10:21] != b'hello world'
    qpair.waitdone(2)
    assert read_buf[10:21] == b'hello world'

    
def test_write_invalid_nsid(nvme0, nvme0n1, cq, sq):
    # first cmd, invalid namespace
    cmd = SQE(1, 0xff)
    buf = PRP(512)
    cmd.prp1 = buf
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x000b

    
def test_write_invalid_nlb(nvme0, nvme0n1, cq, sq):
    ncap = nvme0n1.id_data(15, 8)
    mdts = nvme0.mdts
    
    # first cmd, invalid namespace
    cmd = SQE(1, 1)
    buf = PRP(512)
    cmd.prp1 = buf
    cmd[12] = mdts//512
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x0002  # invalid field

    
def test_write_invalid_nsid_lba(nvme0, nvme0n1, cq, sq):
    ncap = nvme0n1.id_data(15, 8)
    mdts = nvme0.mdts
    
    # first cmd, invalid namespace
    cmd = SQE(1, 0xff)
    buf = PRP(512)
    cmd.prp1 = buf
    cmd[10] = ncap&0xffffffff
    cmd[11] = ncap>>32
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x000b  # invalid namespace or format

    
def test_write_invalid_prp_address_offset(nvme0, cq, sq):
    # first cmd, invalid prp offset
    cmd = SQE(1, 1)
    buf = PRP(4096)
    buf.offset = 1
    cmd.prp1 = buf
    cmd[10] = 1
    cmd[11] = 0
    cmd[12] = 1
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0x0013  # invalid namespace or format




# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-



import time
import pytest
import logging

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem
from scripts.psd import IOCQ, IOSQ, PRP, PRPList, SQE, CQE


@pytest.fixture()
def cq(nvme0):
    ret = IOCQ(nvme0, 1, 10, PRP())
    yield ret
    ret.delete()

@pytest.fixture()
def sq(nvme0, cq):
    ret = IOSQ(nvme0, 1, 10, PRP(), cq.id)
    yield ret
    ret.delete()


def test_write_uncorrectable_large_lba(nvme0, nvme0n1, buf, qpair):
    if not nvme0n1.supports(0x4):
        pytest.skip("Write Uncorrectable is not supported")

    ncap = nvme0n1.id_data(15, 8)
    
    nvme0n1.write_uncorrectable(qpair, ncap-1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write_uncorrectable(qpair, ncap).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write_uncorrectable(qpair, ncap+1).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write_uncorrectable(qpair, ncap-1, 2).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 00/80"):
        nvme0n1.write_uncorrectable(qpair, 0xffffffff00000000).waitdone()


@pytest.mark.parametrize("repeat", range(32))
def test_deallocate_after_write_uncorrectable(nvme0, nvme0n1, repeat, qpair, 
                                              lba_start=0, lba_step=8, lba_count=8):
    if not nvme0n1.supports(0x4):
        pytest.skip("Write Uncorrectable is not supported")

    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supported")


    buf = Buffer(4096)
    pattern = repeat + (repeat<<8) + (repeat<<16) + (repeat<<24)
    write_buf = Buffer(4096, "write", pattern, 32)

    nvme0n1.write_uncorrectable(qpair, lba_start+repeat*lba_step, lba_count).waitdone()
    buf.set_dsm_range(0, lba_start+repeat*lba_step, lba_count)
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    nvme0n1.write(qpair, write_buf, lba_start+repeat*lba_step, lba_count).waitdone()

    
@pytest.mark.parametrize("repeat", range(32))
def test_deallocate_before_write_uncorrectable(nvme0, nvme0n1, repeat, qpair, 
                                              lba_start=0, lba_step=8, lba_count=8):
    if not nvme0n1.supports(0x4):
        pytest.skip("Write Uncorrectable is not supported")

    if not nvme0n1.supports(0x9):
        pytest.skip("dsm is not supported")

    buf = Buffer(4096)
    pattern = repeat + (repeat<<8) + (repeat<<16) + (repeat<<24)
    write_buf = Buffer(4096, "write", pattern, 32)
    q = Qpair(nvme0, 8)

    buf.set_dsm_range(0, lba_start+repeat*lba_step, lba_count)
    nvme0n1.dsm(qpair, buf, 1).waitdone()
    nvme0n1.write_uncorrectable(qpair, lba_start+repeat*lba_step, lba_count).waitdone()
    nvme0n1.write(qpair, write_buf, lba_start+repeat*lba_step, lba_count).waitdone()

    
@pytest.mark.parametrize("repeat", range(32))
def test_write_uncorrectable_read(nvme0, nvme0n1, repeat, qpair, 
                                  lba_start=0, lba_step=8, lba_count=8):
    if not nvme0n1.supports(0x4):
        pytest.skip("Write Uncorrectable is not supported")

    buf = Buffer(4096)
    read_buf = Buffer(4096, "read")
    pattern = repeat + (repeat<<8) + (repeat<<16) + (repeat<<24)
    write_buf = Buffer(4096, "write", pattern, 32)

    nvme0n1.write_uncorrectable(qpair, lba_start+repeat*lba_step, lba_count).waitdone()
    with pytest.warns(UserWarning, match="ERROR status: 02/81"):
        nvme0n1.read(qpair, read_buf, lba_start+repeat*lba_step, lba_count).waitdone()
    nvme0n1.write(qpair, write_buf, lba_start+repeat*lba_step, lba_count).waitdone()
    nvme0n1.read(qpair, read_buf, lba_start+repeat*lba_step, lba_count).waitdone()
    for i in range(lba_count):
        assert read_buf[i*512 + 10] == repeat


def test_write_uncorrectable_invalid_nlb(nvme0, nvme0n1, cq, sq):
    if not nvme0n1.supports(0x4):
        pytest.skip("Write Uncorrectable is not supported")

    ncap = nvme0n1.id_data(15, 8)
    mdts = nvme0.mdts
    
    # first cmd, invalid namespace
    cmd = SQE(4, 1)
    cmd[12] = mdts//512
    sq[0] = cmd
    sq.tail = 1
    time.sleep(0.1)
    status = (cq[0][3]>>17)&0x7ff
    assert status == 0



# Copyright (C) 2020 Crane Chu <cranechu@gmail.com>
# This file is part of pynvme's conformance test
#
# pynvme's conformance test is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# pynvme's conformance test is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pynvme's conformance test. If not, see
# <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-


import time
import pytest
import logging
import warnings

from nvme import Controller, Namespace, Buffer, Qpair, Pcie, Subsystem


def test_aer_limit_exceeded(nvme0):
    aerl = nvme0.id_data(259)+1
    logging.info(aerl)
    
    # another one is sent in defaul nvme init
    for i in range(aerl-1):
        nvme0.aer()

    # send one more
    with pytest.warns(UserWarning, match="ERROR status: 01/05"):
        nvme0.aer()
        nvme0.getfeatures(7).waitdone()

    # abort all
    with pytest.warns(UserWarning, match="ERROR status: 00/07"):
        for i in range(100):
            nvme0.abort(127-i).waitdone()


def test_aer_sanitize(pcie, buf):
    # aer callback function
    aer_cdw0 = 0
    def aer_cb(cdw0, status):
        nonlocal aer_cdw0; aer_cdw0 = cdw0
        
    def nvme_init(nvme0):
        # 2. disable cc.en and wait csts.rdy to 0
        nvme0[0x14] = 0
        while not (nvme0[0x1c]&0x1) == 0: pass

        # 3. set admin queue registers
        nvme0.init_adminq()

        # 4. set register cc
        nvme0[0x14] = 0x00460000

        # 5. enable cc.en
        nvme0[0x14] = 0x00460001

        # 6. wait csts.rdy to 1
        while not (nvme0[0x1c]&0x1) == 1: pass

        # 7. identify controller
        nvme0.identify(Buffer(4096)).waitdone()

        # 8. create and identify all namespace
        nvme0.init_ns()

        # 9. set/get num of queues
        nvme0.setfeatures(0x7, cdw11=0x00ff00ff).waitdone()
        nvme0.getfeatures(0x7).waitdone()

        # 10. send out all aer
        aerl = nvme0.id_data(259)+1
        for i in range(aerl):
            nvme0.aer(cb=aer_cb)

    nvme0 = Controller(pcie, nvme_init_func=nvme_init)
    if nvme0.id_data(331, 328) == 0:
        pytest.skip("sanitize operation is not supported")
            
    logging.info("supported sanitize operation: %d" % nvme0.id_data(331, 328))
    nvme0.sanitize().waitdone()  # sanitize command is completed

    # check sanitize status in log page
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()
        while buf.data(3, 2) & 0x7 != 1:  # sanitize operation is not completed
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)

    # check sanitize status
    nvme0.getlogpage(0x81, buf, 20).waitdone()
    assert buf.data(3, 2) & 0x7 == 1
        
    # test sanitize once more with new aer
    nvme0.sanitize().waitdone()  # sanitize command is completed
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x81, buf, 20).waitdone()
        while buf.data(3, 2) & 0x7 != 1:  # sanitize operation is not completed
            time.sleep(1)
            nvme0.getlogpage(0x81, buf, 20).waitdone()  #L20
            progress = buf.data(1, 0)*100//0xffff
            logging.info("%d%%" % progress)
    assert aer_cdw0 == 0x810106


def test_aer_mask_event(nvme0):
    config = nvme0.getfeatures(0xb).waitdone()
    assert config&1

    # disable the SMART/health event
    nvme0.setfeatures(0xb, cdw11=config&0xfffffffe).waitdone()
    config = nvme0.getfeatures(0xb).waitdone()
    assert not config&1

    # set temperature to generate event
    smart_log = Buffer()
    
    nvme0.getlogpage(0x02, smart_log, 512).waitdone()
    ktemp = smart_log.data(2, 1)
    logging.info("temperature: %d degreeF" % ktemp)

    # over composite temperature threshold
    nvme0.setfeatures(4, cdw11=ktemp-10).waitdone()

    # AER should be triggered here
    with pytest.warns(UserWarning, match="AER notification is triggered"):
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()
        nvme0.setfeatures(4, cdw11=ktemp-20).waitdone()
        nvme0.getlogpage(0x02, smart_log, 512).waitdone()    
        assert smart_log.data(0) & 0x2
    
        # revert to default
        orig_config = nvme0.getfeatures(4, sel=1).waitdone()
        nvme0.setfeatures(4, cdw11=orig_config).waitdone()
        
    nvme0.setfeatures(0xb, cdw11=config).waitdone()
    
